from __future__ import annotations
from typing import TYPE_CHECKING

import re

from trueseeing.core.android.model import Op, Token

if TYPE_CHECKING:
  from typing import Iterator, TypeVar

  T = TypeVar('T')

class OpAnalyzer:
  def get_insn(self, o: Op) -> str:
    t = self.get_mnemonic(o)
    if t.t != 'id':
      raise ValueError(t)
    return t.v

  def get_mnemonic(self, o: Op) -> Token:
    return self._first(self._lex(o.l))

  def get_param(self, o: Op, i: int) -> Token:
    return self._getn(self._lex(o.l), i + 1)

  def get_param_count(self, o: Op) -> int:
    return self._count(self._lex(o.l)) - 1

  def tokenize(self, o: Op) -> Iterator[Token]:
    return self._lex(o.l)

  def _lex(cls, l: str) -> Iterator[Token]:
    for m in re.finditer(r':(?P<label>[a-z0-9_-]+)|{\s*(?P<multilabel>(?::[a-z0-9_-]+(?: .. )*)+\s*)}|\.(?P<directive>[a-z0-9_-]+)|"(?P<string>.*)"|#(?P<comment>.*)|(?P<reg>[vp][0-9]+)|{(?P<multireg>[vp0-9,. ]+)}|(?P<id>[a-z][a-z/-]*[a-z0-9/-]*)|(?P<reflike>[^ ]+)', l):
      key = m.lastgroup
      if key:
        value = m.group(key)
        if key == 'reflike' and value == ',':
          pass
        else:
          yield Token(key, value)

  def _first(self, xs: Iterator[T]) -> T:
    return self._getn(xs, 0)

  def _getn(self, xs: Iterator[T], i: int) -> T:
    for n, x in enumerate(xs):
      if n == i:
        return x
    raise IndexError()

  def _count(self, xs: Iterator[T]) -> int:
    n = 0
    for _ in xs:
      n += 1
    return n



from __future__ import annotations
from typing import TYPE_CHECKING

from trueseeing.core.db import Query, StorePrep
from trueseeing.core.z import zd

if TYPE_CHECKING:
  from typing import Optional, Iterator, Iterable, Tuple
  from trueseeing.core.ios.model import Call

class IPAStorePrep(StorePrep):
  def stage1(self) -> None:
    super().stage1()
    from importlib.resources import files
    self.c.executescript((files('trueseeing')/'libs'/'ios'/'store.0.sql').read_text())

  def stage2(self) -> None:
    from importlib.resources import files
    self.c.executescript((files('trueseeing')/'libs'/'ios'/'store.1.sql').read_text())

  def _get_cache_schema_id(self) -> int:
    return super()._get_cache_schema_id() ^ 0x00f28883

class IPAQuery(Query):
  def file_enum(self, pat: Optional[str], patched: bool = False, regex: bool = False, neg: bool = False) -> Iterable[Tuple[str, bytes]]:
    if pat is not None:
      stmt0 = 'select path, z, blob from files where {neg} path {op} :pat'.format(neg='not' if neg else '', op=('like' if not regex else 'regexp'))
      stmt1 = 'select path, A.z as z, coalesce(B.blob, A.blob) as blob from files as A full outer join patches as B using (path) where path {op} :pat'.format(op=('like' if not regex else 'regexp'))
      for n, z, o in self.db.execute(stmt1 if patched else stmt0, dict(pat=pat)):
        yield n, zd(o) if z else o
    else:
      stmt2 = 'select path, z, blob from files'
      stmt3 = 'select path, A.z as z, coalesce(B.blob, A.blob) from files as A full outer join patches as B using (path)'
      for n, z, o in self.db.execute(stmt3 if patched else stmt2):
        yield n, zd(o) if z else o

  def call_add_batch(self, gen: Iterator[Call]) -> None:
    stmt0 = 'insert into ncalls (priv, swift, cpp, objc, target, path, sect, offs) values (:priv, :swift, :cpp, :objc, :target, :path, :sect, :offs)'
    self.db.executemany(stmt0, gen)

  def call_count(self) -> int:
    stmt0 = 'select count(1) from ncalls'
    for n, in self.db.execute(stmt0):
      return n # type:ignore[no-any-return]
    return 0

  def calls(self, priv: bool = False, api: bool = False) -> Iterator[Call]:
    stmt0 = 'select priv, swift, cpp, objc, target, path, sect, offs from ncalls'
    stmt1 = 'select priv, swift, cpp, objc, target, path, sect, offs from ncalls where priv=:is_priv'
    for priv, swift, cpp, objc, target, path, sect, offs in self.db.execute(stmt1 if (priv or api) else stmt0, dict(is_priv=priv)):
      yield dict(
        path=path,
        sect=sect,
        offs=offs,
        priv=priv,
        swift=swift,
        objc=objc,
        cpp=cpp,
        target=target,
      )



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass

class FatalError(Exception):
    pass

class InvalidSchemaError(Exception):
    pass

class InvalidContextError(Exception):
    pass

class InvalidFileFormatError(Exception):
    pass

class InvalidConfigKeyError(Exception):
    pass



from __future__ import annotations
from typing import TYPE_CHECKING

import itertools
import re

from trueseeing.core.android.model import SignatureMixin

if TYPE_CHECKING:
  from trueseeing.api import Signature, SignatureHelper, SignatureMap

class ManifestOpenPermissionDetector(SignatureMixin):
  _id = 'manifest-open-permission'
  _cvss = 'CVSS:3.0/AV:P/AC:H/PR:N/UI:R/S:U/C:N/I:N/A:N/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestOpenPermissionDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects declarated permissions')}

  async def detect(self) -> None:
    context = self._get_context()

    # TBD: compare with actual permission needs
    for p in context.permissions_declared():
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cfd='certain',
        cvss=self._cvss,
        title='open permissions',
        info0=p,
        aff0='AndroidManifest.xml',
        summary="Application is requesting one or more permissions.",
        desc="Application is requesting one or more permissions.  Permissions are an important security system of Android.  They control accesses to sensitive information (e.g. GPS, IMEI/IMSI, process stats, accounts, contacts, SMSs) or possibly dangerous/costly operation (e.g. SMSs, internet access, controlling system services, obstructing screens.)  Requesting ones are vital for proper functioning of application, though abusage leads to hurt privacy or device availability.  This issue is just an observation; requesting permissions alone does not constitute an security issue.",
      ))

class ComponentNamePolicy:
  def __init__(self) -> None:
    from importlib.resources import files
    with (files('trueseeing')/'libs'/'tlds.txt').open('r', encoding='utf-8') as f:
      self._re_tlds = re.compile('^(?:{})$'.format('|'.join(re.escape(l.strip()) for l in f if l and not l.startswith('#'))), flags=re.IGNORECASE)

  def looks_public(self, name: str) -> bool:
    if '.' in name:
      gtld = name.split('.')[0]
      return (gtld == 'android') or ('.intent.action.' in name) or bool(self._re_tlds.search(gtld))
    else:
      return False

class ManifestManipActivity(SignatureMixin):
  _id = 'manifest-manip-activity'
  description = 'Detects exported Activity'
  _cvss1 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _cvss2 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:L/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestManipActivity(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects exported Activity')}

  async def detect(self) -> None:
    context = self._get_context()
    policy = ComponentNamePolicy()
    ns = dict(android='http://schemas.android.com/apk/res/android')

    for name in set(itertools.chain(
        context.parsed_manifest().xpath('//activity[not(@android:permission)]/intent-filter/../@android:name', namespaces=ns),
        context.parsed_manifest().xpath('//activity[not(@android:permission) and (@android:exported="true")]/@android:name', namespaces=ns),
    )):
      filter_ = [name for name in context.parsed_manifest().xpath(f'//activity[@android:name="{name}"]/intent-filter/action/@android:name', namespaces=ns) if not policy.looks_public(name)]
      if not filter_:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss1,
          title='manipulatable Activity',
          info0=name,
          aff0='AndroidManifest.xml',
          summary="Application is exporting one or more activities.",
          desc="Application is exporting one or more activities.  Activities are entrypoints to the application.  Exporting enables them to be invoked from other applications or system.  Unnecessary export increases attack surfaces.  Please note that Android automatically exports ones with IntentFilter defined in the manifest.  This issue is just an observation; exporting activities alone does not constitute an security issue.",
          sol="Review them, and restrict access with application-specific permissions if necessary."
        ))
      else:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss2,
          title='manipulatable Activity with private action names',
          info0=name,
          info1=', '.join(filter_),
          aff0='AndroidManifest.xml',
          summary="Application is exporting one or more activities using seemingly private action names, suggesting inadvent export.",
          desc="Application is exporting one or more activities using seemingly private action names, suggesting inadvent export.  Activities are entrypoints to the application.  Exporting enables them to be invoked from other applications or system.  Inadvent exporting enables malwares or malicious users to manipulate the application.  Please note that Android automatically exports ones with IntentFilter defined in the manifest.",
          sol="Review them, and restrict access with application-specific permissions if necessary."
        ))

class ManifestManipBroadcastReceiver(SignatureMixin):
  _id = 'manifest-manip-broadcastreceiver'
  _cvss1 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _cvss2 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:L/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestManipBroadcastReceiver(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects exported BroadcastReceiver')}

  async def detect(self) -> None:
    context = self._get_context()
    policy = ComponentNamePolicy()
    ns = dict(android='http://schemas.android.com/apk/res/android')

    # FIXME: catch API < 9
    for name in set(itertools.chain(
        context.parsed_manifest().xpath('//receiver[not(@android:permission) and not(@android:exported="false")]/intent-filter/../@android:name', namespaces=ns),
        context.parsed_manifest().xpath('//receiver[not(@android:permission) and (@android:exported="true")]/@android:name', namespaces=ns),
    )):
      filter_ = [name for name in context.parsed_manifest().xpath(f'//receiver[@android:name="{name}"]/intent-filter/action/@android:name', namespaces=ns) if not policy.looks_public(name)]
      if not filter_:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss1,
          title='manipulatable BroadcastReceiver',
          info0=name,
          aff0='AndroidManifest.xml',
          summary="Application is exporting one or more broadcast receivers.",
          desc="Application is exporting one or more broadcast receivers.  Broadcast receivers are system-wide event listeners of the application.  Exporting enables them to be invoked from other applications or system.  Unnecessary export increases attack surfaces.  Please note that Android automatically exports ones with IntentFilter defined in the manifest.  This issue is just an observation; exporting broadcast receivers alone does not constitute an security issue.",
          sol="Review them and restrict access with application-specific permissions if necessary.  Consider the use of LocalBroadcastReceiver for ones that system-wide reachability is not needed."
        ))
      else:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss2,
          title='manipulatable BroadcastReceiver with private action names',
          info0=name,
          info1=', '.join(filter_),
          aff0='AndroidManifest.xml',
          summary="Application is exporting one or more broadcast receivers using seemingly private action names, suggesting inadvent export.",
          desc="Application is exporting one or more broadcast receivers using seemingly private action names, suggesting inadvent export.  Broadcast receivers are system-wide event listeners of the application.  Exporting enables them to be invoked from other applications or system.  Inadvent exporting enables malwares or malicious users to manipulate the application.  Please note that Android automatically exports ones with IntentFilter defined in the manifest.",
          sol="Review them, and restrict access with application-specific permissions if necessary."
        ))

class ManifestManipContentProvider(SignatureMixin):
  _id = 'manifest-manip-contentprovider'
  _cvss1 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _cvss2 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:L/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestManipContentProvider(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects exported ContentProvider')}

  async def detect(self) -> None:
    context = self._get_context()
    policy = ComponentNamePolicy()
    ns = dict(android='http://schemas.android.com/apk/res/android')

    for name in set(itertools.chain(
        context.parsed_manifest().xpath('//provider[not(@android:permission)]/intent-filter/../@android:name', namespaces=dict(android='http://schemas.android.com/apk/res/android')),
        context.parsed_manifest().xpath('//provider[not(@android:permission) and (@android:exported="true")]/@android:name', namespaces=dict(android='http://schemas.android.com/apk/res/android')),
    )):
      filter_ = [name for name in context.parsed_manifest().xpath(f'//receiver[@android:name="{name}"]/intent-filter/action/@android:name', namespaces=ns) if not policy.looks_public(name)]
      if not filter_:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss1,
          title='manipulatable ContentProvider',
          info0=name,
          aff0='AndroidManifest.xml',
          summary="Application is exporting one or more content providers.",
          desc="Application is exporting one or more content providers.  Content providers defines REST/RDBMS-like IPC mechanism for the application.  Exporting enables them to be invoked from other applications or system.  Unnecessary export increases attack surfaces.  Please note that Android automatically exports them (API 8 or ealier) or ones with IntentFilter defined in the manifest (API level 9 or later).  This issue is just an observation; exporting content providers alone does not constitute an security issue.",
          sol='''\
  Review them and explicitly unexport or restrict access with application-specific permissions if necessary.  To explicitly unexporting an content provider, define the following attribute to the <provider> tag in the manifest:

  android:export="false"
  '''
        ))
      else:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss2,
          title='manipulatable ContentProvider with private action names',
          info0=name,
          info1=', '.join(filter_),
          aff0='AndroidManifest.xml',
          summary="Application is exporting one or more content providers using seemingly private action names, suggesting inadvent export.",
          desc="Application is exporting one or more content providers using seemingly private action names, suggesting inadvent export.  Content providers defines REST/RDBMS-like IPC mechanism for the application.  Exporting enables them to be invoked from other applications or system.  Inadvent exporting enables malwares or malicious users to manipulate the application.  Please note that Android automatically exports them (API 8 or ealier) or ones with IntentFilter defined in the manifest (API level 9 or later).",
          sol='''\
  Review them and explicitly unexport or restrict access with application-specific permissions if necessary.  To explicitly unexporting an content provider, define the following attribute to the <provider> tag in the manifest:

  android:export="false"
  '''
        ))

class ManifestManipBackup(SignatureMixin):
  _id = 'manifest-manip-backup'
  _cvss = 'CVSS:3.0/AV:A/AC:L/PR:N/UI:R/S:U/C:H/I:H/A:H/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestManipBackup(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects enabled backup bit')}

  async def detect(self) -> None:
    context = self._get_context()
    manif = context.parsed_manifest()
    for e in manif.xpath('//application[not(@android:allowBackup="false")]', namespaces=dict(android='http://schemas.android.com/apk/res/android')):
      if min(context.get_target_sdk_version(), context.get_min_sdk_version()) < 31:
        fbc_exists = (e.attrib.get('{{{ns}}}fullBackupContent'.format(ns='http://schemas.android.com/apk/res/android')) is not None)
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain' if not fbc_exists else 'tentative',
          cvss=self._cvss,
          title='manipulatable backups',
          aff0='AndroidManifest.xml',
          summary="Application data can be backed up and restored with the Auto Backup feature.",
          desc="Application data can be backed up and restored with the Auto Backup feature, thusly making it subjectible to the backup attack.",
          sol='''\
Review them, and consider controlling backupable data with the Auto Backup feature or completely opt-out from it if necessary.  To control backupable data, associate XML with the following attribute to the <application> tag.

android:fullBackupContent="@xml/fbc"

(Please refer https://developer.android.com/guide/topics/data/autobackup#XMLSyntax for details)

To opt-out, define the following attribute to the <application> tag in the manifest:

android:allowBackup="false"
'''
        ))

class ManifestDebuggable(SignatureMixin):
  _id = 'manifest-debuggable'
  _cvss = 'CVSS:3.0/AV:A/AC:L/PR:N/UI:N/S:U/C:H/I:H/A:H/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestDebuggable(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects enabled debug bits')}

  async def detect(self) -> None:
    context = self._get_context()
    if context.parsed_manifest().xpath('//application[@android:debuggable="true"]', namespaces=dict(android='http://schemas.android.com/apk/res/android')):
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cfd='certain',
        cvss=self._cvss,
        title='app is debuggable',
        aff0='AndroidManifest.xml',
        summary="Application can be debugged.",
        desc="Application can be debugged (the debuggable bit is set.)  Debugging it gives attackers complete control of its process memory and control flow.",
        sol='''\
Disable the debuggable bit.  To disable it, define the following attribute to the <application> tag in the manifest:

android:debuggable="false"
'''
      ))

class ManifestCleartextPermitted(SignatureMixin):
  _id = 'manifest-cleartext-permitted'
  _cvss = 'CVSS:3.0/AV:A/AC:H/PR:N/UI:N/S:U/C:L/I:L/A:N/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ManifestCleartextPermitted(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects usesCleartextTraffic flag')}

  async def detect(self) -> None:
    context = self._get_context()
    manif = context.parsed_manifest()
    api_level = context.get_min_sdk_version()
    if api_level < 24:
      if not manif.xpath('//application[@android:usesCleartextTraffic="false"]', namespaces=dict(android='http://schemas.android.com/apk/res/android')):
        self._raise('AndroidManifest.xml')
    else:
      for e in manif.xpath('//application', namespaces=dict(android='http://schemas.android.com/apk/res/android')):
        if e.attrib.get('{{{ns}}}networkSecurityConfig'.format(ns='http://schemas.android.com/apk/res/android')) is None:
          if e.attrib.get('{{{ns}}}usesCleartextTraffic'.format(ns='http://schemas.android.com/apk/res/android'), 'true' if api_level < 28 else 'false').lower == 'true':
            self._raise('AndroidManifest.xml')
            break
      for fn, xp in context.xml_resources():
        if 'network-security-config' in xp.tag.lower():
          if api_level < 28:
            if not xp.xpath('.//*[@cleartextTrafficPermitted="false"]'):
              self._raise(context.source_name_of_disassembled_resource(fn))
          else:
            if xp.xpath('.//*[@cleartextTrafficPermitted="true"]'):
              self._raise(context.source_name_of_disassembled_resource(fn))

  def _raise(self, path: str) -> None:
    self._helper.raise_issue(self._helper.build_issue(
      sigid=self._id,
      cfd='certain',
      cvss=self._cvss,
      title='cleartext traffic is permitted',
      aff0=path,
      summary="Application can use cleartext traffic.",
      desc="Application can use cleartext traffic.  Cleartext traffic are prone to be intercept and/or manipulated by an active attacker on the wire.",
      sol='''\
Consider migrating all the servers on TLS and disable use of cleartext traffic.  To disable one, associate an Network Security Config disallowing one (API > 24), or discourage ones use in <application> tag as follows (API > 23):

<application android:useCleartextTraffic="false">
'''
    ))



from __future__ import annotations
from typing import TYPE_CHECKING

import asyncio
from functools import cache
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Any, Optional, TypeVar, Iterator, AsyncIterator, Type, FrozenSet, Dict, Set, Union, SupportsIndex
  from typing_extensions import Buffer
  T = TypeVar('T')

def noneif(x: Any, defaulter: Any) -> Any:
  if x is not None:
    return x
  else:
    if callable(defaulter):
      return defaulter()
    else:
      return defaulter

def _check_return_code(p: Any, args: Any, out: Any, err: Any) -> None:
  code: int
  if isinstance(p, int):
    code = p
  elif hasattr(p, 'returncode'):
    code = p.returncode
  if code:
    from subprocess import CalledProcessError
    raise CalledProcessError(code, args, out, err)

@cache
def require_in_path(cmd: str, cmdline: str) -> None:
  from subprocess import run, CalledProcessError
  try:
    run(cmdline, capture_output=True, check=True, shell=True)
  except CalledProcessError:
    ui.fatal('not found: {cmd}')

async def invoke(as_: str, redir_stderr: bool = False, catch_stderr: bool = False) -> str:
  from asyncio import create_subprocess_shell
  from subprocess import PIPE, STDOUT
  p = await create_subprocess_shell(as_, stdout=PIPE, stderr=(STDOUT if redir_stderr else (PIPE if catch_stderr else None)))
  out, err = await p.communicate()
  _check_return_code(p, as_, out, err)
  return out.decode('UTF-8')

async def invoke_passthru(as_: str, nocheck: bool = False) -> None:
  from asyncio import create_subprocess_shell
  p = await create_subprocess_shell(as_)
  await p.communicate()
  if not nocheck:
    _check_return_code(p, as_, None, None)

class _UniversalBufferPatch(bytearray):
  def find(self, __sub: Union[Buffer, SupportsIndex], __start: Optional[SupportsIndex] = None, __end: Optional[SupportsIndex] = None) -> int:
    return super().replace(b'\r', b'\n').find(__sub, __start, __end)

async def invoke_streaming(as_: str, redir_stderr: bool = False) -> AsyncIterator[bytes]:
  from subprocess import PIPE, STDOUT
  p = await asyncio.create_subprocess_shell(as_, stdout=PIPE, stderr=(STDOUT if redir_stderr else None))
  p.stdout._buffer = _UniversalBufferPatch(p.stdout._buffer) # type: ignore[union-attr]
  try:
    l: Optional[bytes] = None
    if p.stdout is not None:
      async for l in p.stdout:
        yield l
    _check_return_code(await p.wait(), as_, l, None)
  finally:
    if p.returncode is None:
      try:
        t = asyncio.create_task(p.wait())
        await asyncio.wait([t], timeout=3., return_when=asyncio.ALL_COMPLETED)
      except asyncio.TimeoutError:
        ui.warn('process does not seem to terminate, killing it')
        p.kill()
        await p.wait()

def invoke_sync(as_: str, redir_stderr: bool = False, catch_stderr: bool = False) -> str:
  from subprocess import PIPE, STDOUT, run
  p = run(as_, shell=True, stdout=PIPE, stderr=(STDOUT if redir_stderr else (PIPE if catch_stderr else None)))
  out, err = p.stdout, p.stderr
  _check_return_code(p, as_, out, err)
  return out.decode('UTF-8')

async def try_invoke(as_: str) -> Optional[str]:
  from subprocess import CalledProcessError
  try:
    return await invoke(as_)
  except CalledProcessError:
    return None

def copytree(src: str, dst: str, divisor: Optional[int] = 256) -> Iterator[int]:
  import os
  from shutil import copy2, copystat

  nr = 0
  for sp, dns, fns in os.walk(src):
    dp = os.path.realpath(os.path.join(dst, os.path.relpath(sp, src)))
    for dn in dns:
      os.makedirs(os.path.join(dp, dn), exist_ok=True)
      copystat(os.path.join(sp, dn), os.path.join(dp, dn))
    for fn in fns:
      copy2(os.path.join(sp, fn), os.path.join(dp, fn))
      if divisor is None or (nr % divisor == 0):
        yield nr
      nr += 1

def copy_from_pack(src: str, dst: str, prefix: str, divisor: Optional[int] = 256) -> Iterator[int]:
  import tarfile

  nr = 0
  with tarfile.open(src) as tf:
    for name in tf.getnames():
      if name.startswith(prefix):
        tf.extract(name, dst)
      if divisor is None or (nr % divisor == 0):
        yield nr
      nr += 1

def move_as_output(src: str, dst: str, divisor: Optional[int] = 256, allow_orphans: bool = False) -> Iterator[int]:
  import os
  from trueseeing.core.env import is_in_container
  if not is_in_container():
    try:
      os.rename(src, dst)
      yield 0
      return
    except OSError:
      pass

  nr = 0

  from shutil import copy2, copystat
  for sp, dns, fns in os.walk(src, topdown=False):
    dp = os.path.realpath(os.path.join(dst, os.path.relpath(sp, src)))
    for dn in dns:
      copystat(os.path.join(sp, dn), os.path.join(dp, dn))
      if not allow_orphans:
        os.rmdir(os.path.join(sp, dn))
    for fn in fns:
      os.makedirs(dp, exist_ok=True)
      copy2(os.path.join(sp, fn), os.path.join(dp, fn))
      if not allow_orphans:
        os.remove(os.path.join(sp, fn))
      if divisor is None or (nr % divisor == 0):
        yield nr
      nr += 1

def pack_as_output(src: str, dst: str, prefix: str, subformat: str, divisor: Optional[int] = 256, allow_orphans: bool = False) -> Iterator[int]:
  import os

  nr = 0

  import tarfile
  kwargs: Dict[str, int] = dict()
  if subformat in ['gz']:
    kwargs.update(dict(compresslevel=3))

  with tarfile.open(dst, 'w:{}'.format(subformat), **kwargs) as tf:  # type: ignore[call-overload]
    for sp, dns, fns in os.walk(src, topdown=False):
      dp = os.path.join(prefix, os.path.relpath(sp, src))
      for dn in dns:
        tf.add(os.path.join(sp, dn), arcname=os.path.join(dp, dn), recursive=False)
        if not allow_orphans:
          os.rmdir(os.path.join(sp, dn))
      for fn in fns:
        tf.add(os.path.join(sp, fn), arcname=os.path.join(dp, fn), recursive=False)
        if not allow_orphans:
          os.remove(os.path.join(sp, fn))
        if divisor is None or (nr % divisor == 0):
          yield nr
        nr += 1

def get_public_subclasses(mod: Any, typ: Type[T], nopat: Optional[str] = None) -> Iterator[Type[T]]:
  import re
  from inspect import getmembers, isclass
  for n, clazz in getmembers(mod, lambda x: isclass(x) and x != typ and issubclass(x, typ)):
    if not n.startswith('_') and not (nopat and re.search(nopat, n)):
      yield clazz

def get_missing_methods(clazz: Any) -> FrozenSet[str]:
  from inspect import isclass
  assert isclass(clazz)
  return getattr(clazz, '__abstractmethods__', frozenset())

def get_fully_qualified_classname(clazz: Any) -> str:
  return '.'.join([clazz.__module__, clazz.__name__])

def discover_modules_under(anchor: str) -> Iterator[str]:
  import os.path
  import re
  from importlib.resources import files
  from glob import iglob
  seen: Set[str] = set()
  basepath = str(files(anchor))
  for path in iglob(os.path.join(basepath, '**', '*.p*'), recursive=True):
    if '__' in path or os.path.basename(path).startswith('_') or not re.search(r'\.pyc?$', path):
      continue
    n = '{}.{}'.format(anchor, re.sub(r'\.pyc?$', '', os.path.relpath(path, basepath)).replace('/', '.'))
    if n not in seen:
      seen.add(n)
      yield n



from __future__ import annotations
from typing import TYPE_CHECKING

import asyncio
import re
import math
from functools import cache

from trueseeing.core.android.model import InvocationPattern
from trueseeing.core.android.model import SignatureMixin
from trueseeing.core.android.analyze.flow import DataFlow

if TYPE_CHECKING:
  from typing import Dict, Iterable, Optional, Any
  from trueseeing.core.android.store import APKStore
  from trueseeing.core.android.db import APKQuery
  from trueseeing.core.android.model import Op
  from trueseeing.api import Signature, SignatureHelper, SignatureMap

class CryptoStaticKeyDetector(SignatureMixin):
  _id = 'crypto-static-keys'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:C/C:L/I:L/A:N/'
  _cvss_nonkey = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _cvss_pubkey = 'CVSS:3.0/AV:P/AC:H/PR:N/UI:N/S:C/C:N/I:L/A:N/'
  _cvss_privkey = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:C/C:H/I:H/A:N/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return CryptoStaticKeyDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id: dict(e=self.detect, d='Detects cryptographic function usage with static keys')}

  @classmethod
  def _entropy_of(cls, string: str) -> float:
    o = 0.0
    m: Dict[str, int] = dict()
    for c in string:
      m[c] = m.get(c, 0) + 1
    for cnt in m.values():
      freq = float(cnt) / len(string)
      o -= freq * (math.log(freq) / math.log(2))
    return o

  @classmethod
  def _assumed_randomness_of(cls, string: str) -> float:
    try:
      return cls._entropy_of(string) / float(math.log(len(string)) / math.log(2))
    except ValueError:
      return 0

  def _important_args_on_invocation(self, k: Op) -> Iterable[int]:
    method_name = self._an.get_param(k, 1).v
    if re.match('L.*/(SecretKey|(Iv|GCM)Parameter|(PKCS8|X509)EncodedKey)Spec-><init>|L.*/MessageDigest;->update', method_name):
      yield 0
    else:
      yield from range(len(DataFlow.decoded_registers_of_set(self._an.get_param(k, 0))))

  async def detect(self) -> None:
    await asyncio.gather(self._do_detect_case1(), self._do_detect_case2())

  async def _do_detect_case1(self) -> None:
    import base64
    import binascii

    def looks_like_real_key(k: str) -> bool:
      # XXX: silly
      return len(k) >= 8 and not any(x in k for x in ('Padding', 'SHA1', 'PBKDF2', 'Hmac', 'emulator'))

    pat_case2 = '^M[C-I][0-9A-Za-z+/=-]{48,}'
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-', '^Ljavax?.*/(SecretKey|(Iv|GCM)Parameter|(PKCS8|X509)EncodedKey)Spec|^Ljavax?.*/MessageDigest;->(update|digest)')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        for nr in self._important_args_on_invocation(cl):
          for found in DataFlow(q).solved_possible_constant_data_in_invocation(cl, nr):
            try:
              if re.search(pat_case2, found):
                continue

              decoded = base64.b64decode(found)
              info0 = '"{target_val}" [{target_val_len}] (base64; "{decoded_val}" [{decoded_val_len}])'.format(target_val=found, target_val_len=len(found), decoded_val=binascii.hexlify(decoded).decode('ascii'), decoded_val_len=len(decoded))
            except (ValueError, binascii.Error):
              info0 = f'"{found}" [{len(found)}]'

            if looks_like_real_key(found):
              self._helper.raise_issue(self._helper.build_issue(
                sigid=self._id,
                cvss=self._cvss,
                title='insecure cryptography: static keys detected',
                info0=info0,
                info1=q.method_call_target_of(cl.addr),
                aff0=qn,
                summary='Traces of cryptographic material has been found the application binary.',
                desc='''\
Traces of cryptographic material has been found in the application binary.  If cryptographic material is hardcoded, attackers can extract or replace them.
''',
                sol='''\
Use a device or installation specific information, or obfuscate them.
'''
              ))
            else:
              self._helper.raise_issue(self._helper.build_issue(
                sigid=self._id,
                cvss=self._cvss_nonkey,
                cfd='tentative',
                title='Cryptographic constants detected',
                info0=info0,
                info1=q.method_call_target_of(cl.addr),
                aff0=qn,
                summary='Possible cryptographic constants have been found.',
                desc='''\
Possible cryptographic constants has been found in the application binary.
'''
              ))
      except IndexError:
        pass

  async def _do_detect_case2(self) -> None:
    # XXX: Crude detection
    def should_be_secret(store: APKStore, k: Op, val: str) -> bool:
      name = store.query().qualname_of(k.addr)
      if name:
        return name.lower() in ['inapp','billing','iab','sku','store','key']
      else:
        return False

    pat = '^M[C-I][0-9A-Za-z+/=-]{48,}'
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.consts(InvocationPattern('const-string', pat)):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      val = self._an.get_param(cl, 1).v
      typ = self._inspect_value_type(val)
      param = self._build_template_params(val, typ)
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cvss=self._cvss_nonkey if param['nonkey'] else (self._cvss_pubkey if not param['private'] else self._cvss_privkey),
        cfd='certain' if typ else ('firm' if should_be_secret(store, cl, val) else 'tentative'),
        title=('insecure cryptography: {key} detected' if not param['nonkey'] else '{key} detected').format(**param),
        info0='"{val}" [{len}] ({keytype})'.format(**param),
        aff0=qn,
        summary='Traces of {key}s have been found the application binary.'.format(**param),
        desc='''\
Traces of {key}s have been found in the application binary.  If they are hardcoded, attackers can extract or replace them.'''.format(**param),
        sol='''\
Use a device or installation specific information, or obfuscate them.
'''
      ))

    for name, val in context.string_resources():
      if re.match(pat, val):
        typ = self._inspect_value_type(val)
        param = self._build_template_params(val, typ)
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cvss=self._cvss_nonkey if param['nonkey'] else (self._cvss_pubkey if not param['private'] else self._cvss_privkey),
          cfd='certain' if typ else 'firm',
          title=('insecure cryptography: {key} detected' if param['nonkey'] else '{key} detected').format(**param),
          info0='"{val}" [{len}] ({keytype})'.format(**param),
          aff0=f'R.string.{name}',
          summary='Traces of {key}s have been found the application binary.'.format(**param),
          desc='''\
Traces of {key}s have been found in the application binary.  If they are hardcoded, attackers can extract or replace them.'''.format(**param),
          sol='''\
Use a device or installation specific information, or obfuscate them.
'''
        ))

  def _build_template_params(self, val: str, typ: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    param = dict(
      nonkey=True,
      val=val,
      len=len(val),
      key='cryptographic material',
      keytype='ASN.1',
    )
    if typ:
      if typ['type'] == 'cert':
        param.update(dict(
          private=False,
          key='X.509 certificate',
          keytype=('{}-bit {} with {}, subject CN: {}, issuer CN: {}{})'.format(typ['bits'], typ['algo'], typ['hashalgo'], typ['subject'], typ['issuer'], ' [self-signed]' if typ['selfsign'] != 'no' else ''))
        ))
      elif typ['type'] == 'pub':
        param.update(dict(
          nonkey=False,
          private=False,
          key='public key',
          keytype=('{}-bit {}'.format(typ['bits'], typ['algo'])),
        ))
      elif typ['type'] == 'priv':
        param.update(dict(
          nonkey=False,
          private=True,
          key='private key',
          keytype=('{}-bit {}'.format(typ['bits'] if isinstance(typ['bits'], int) else '~{:.01f}'.format(typ['bits']), typ['algo'])),
        ))
    return param

  @cache
  def _inspect_value_type(self, v: str) -> Optional[Dict[str, Any]]:
    import binascii
    from base64 import b64decode
    from asn1crypto.x509 import Certificate
    from asn1crypto.keys import PublicKeyInfo, PrivateKeyInfo, RSAPrivateKey
    try:
      r = b64decode(v)
    except binascii.Error as e:
      from trueseeing.core.ui import ui
      ui.warn('inspect_value_type: decode failed ({}); marking as unknown type: {}'.format(e, v))
      return None

    try:
      cert = Certificate.load(r)
      return dict(type='cert', algo=self._read_algo(cert.public_key), hashalgo=self._read_algo(cert.hash_algo), bits=cert.public_key.bit_size, subject=cert.subject.native.get('common_name', '(unknown)'), issuer=cert.issuer.native.get('common_name', '(unknown)'), selfsign=cert.self_signed)
    except ValueError:
      pass
    try:
      pubkey = PublicKeyInfo.load(r)
      return dict(type='pub', algo=self._read_algo(pubkey), bits=pubkey.bit_size)
    except ValueError:
      pass
    try:
      privkey0 = PrivateKeyInfo.load(r)
      return dict(type='priv', algo=self._read_algo(privkey0), bits=privkey0.bit_size)
    except ValueError:
      pass
    try:
      from math import log
      privkey1 = RSAPrivateKey.load(r)
      return dict(type='priv', algo=self._read_algo('rsa'), bits=log(privkey1['modulus'].native) / log(2))
    except ValueError:
      pass
    return None

  @cache
  def _read_algo(self, x: Any) -> str:
    if isinstance(x, str):
      return x.upper()
    else:
      algo = str(x.algorithm)
      if algo != 'ec':
        return algo.upper()
      else:
        if x.curve[0] == 'named':
          return '{} [{}]'.format(algo.upper(), str(x.curve[1]))
        else:
          return '{}'.format(algo.upper())

class CryptoEcbDetector(SignatureMixin):
  _id = 'crypto-ecb'
  _cvss = 'CVSS:3.0/AV:P/AC:H/PR:N/UI:N/S:U/C:L/I:L/A:L/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return CryptoEcbDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id: dict(e=self.detect, d='Detects ECB mode ciphers')}

  async def detect(self) -> None:
    q: APKQuery
    context = self._get_context()
    with context.store().query().scoped() as q:
      for cl in q.invocations(InvocationPattern('invoke-static', r'Ljavax/crypto/Cipher;->getInstance\(Ljava/lang/String;.*?\)')):
        qn = q.qualname_of(cl.addr)
        if context.is_qualname_excluded(qn):
          continue
        try:
          target_val = DataFlow(q).solved_possible_constant_data_in_invocation(cl, 0)
          if any((('ECB' in x or '/' not in x) and 'RSA' not in x) for x in target_val):
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss,
              cfd='certain',
              title='insecure cryptography: cipher might be operating in ECB mode',
              info0=','.join(target_val),
              aff0=qn,
              summary='The application might be using ciphers in ECB mode.',
              desc='''\
              The application might be using symmetric ciphers in ECB mode.  ECB mode is the most basic operating mode that independently transform data blocks.  Indepent transformation leaks information about distribution in plaintext.
  ''',
              sol='''\
  Use CBC or CTR mode.
  '''
            ))
        except DataFlow.NoSuchValueError:
          pass

class CryptoNonRandomXorDetector(SignatureMixin):
  _id = 'crypto-xor'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:C/C:L/I:L/A:L/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return CryptoNonRandomXorDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id: dict(e=self.detect, d='Detects Vernum cipher usage with static keys')}

  async def detect(self) -> None:
    q: APKQuery
    context = self._get_context()
    with context.store().query().scoped() as q:
      for cl in q.ops_of('xor-int/lit8'):
        qn = q.qualname_of(cl.addr)
        if context.is_qualname_excluded(qn):
          continue
        m = re.search(r'([pv][0-9]+), \1, (0x[0-9a-f]+)$', cl.l)
        if m:
          target_val = int(m.group(2), 16)
          if target_val > 1:
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss,
              title='insecure cryptography: non-random XOR cipher',
              info0=f'0x{target_val:02x}',
              aff0=qn
            ))



from __future__ import annotations
from typing import TYPE_CHECKING

from trueseeing.api import Signature

if TYPE_CHECKING:
  from trueseeing.api import SignatureHelper, ConfigMap

class SignatureMixin(Signature):
  def __init__(self, helper: SignatureHelper) -> None:
    self._helper = helper

  def get_configs(self) -> ConfigMap:
    return dict()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from typing import Type, Iterator
  from trueseeing.api import Signature

def discover() -> Iterator[Type[Signature]]:
  from trueseeing.api import Signature
  from importlib import import_module
  from trueseeing.core.tools import get_public_subclasses, get_missing_methods, discover_modules_under

  for mod in discover_modules_under('trueseeing.sig'):
    m = import_module(mod)
    for c in get_public_subclasses(m, Signature, 'SignatureMixin'):  # type:ignore[type-abstract]
      assert not get_missing_methods(c)
      yield c



from __future__ import annotations
from typing import TYPE_CHECKING

import re
from collections import deque
from shlex import shlex

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Dict, Tuple
  from trueseeing.api import CommandHelper, Command, CommandMap, CommandPatternMap

class AliasCommand(CommandMixin):
  _aliases: Dict[str, str]
  _macros: Dict[str, Tuple[int, str, deque[str]]]

  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper
    self._aliases = dict()
    self._macros = dict()

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return AliasCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      '?$?':dict(e=self._help_alias, n='?$?', d='alias help'),
    }

  def get_command_patterns(self) -> CommandPatternMap:
    return {
      r'\$[a-zA-Z0-9=]+':dict(e=self._alias, n='$alias=value', d='alias command'),
      r'\(.+\)':dict(e=self._alias2, raw=True, n='(macro x y; cmd; cmd; ..)', d='define macro'),
      r'\.\(.+\)':dict(e=self._alias2_call, raw=True, n='.(macro x y)', d='call macro'),
      r'^\(\*$':dict(e=self._help_alias2, raw=True, n='(*', d='macro help'),
    }

  async def _help_alias(self, args: deque[str]) -> None:
    if self._aliases:
      ui.success('Aliases:')
      width = 2 + max([len(k) for k in self._aliases.keys()])
      for k in sorted(self._aliases):
        ui.stderr(
          f'${{n:<{width}s}}{{d}}'.format(n=k, d=self._aliases[k])
        )
    else:
      ui.success('no alias known')

  async def _help_alias2(self, line: str) -> None:
    if self._macros:
      ui.success('Macroes:')
      width = 2 + max([len(k) for k in self._macros.keys()])
      for k in sorted(self._macros):
        ui.stderr(
          f'{{n:<{width}s}}{{d}}'.format(n=k, d=self._macros[k][1])
        )
    else:
      ui.success('no macro known')

  async def _alias(self, args: deque[str]) -> None:
    cmd = args.popleft()
    n = cmd[1:]

    if args:
      op = args.popleft()
      if op != '=':
        ui.fatal('alias cannot take arguments')
      if args:
        val = args.popleft()
      else:
        val = None
    else:
      op = None
      val = None

    if op is None:
      try:
        v = self._aliases[n]
      except KeyError:
        ui.error('invalid command, type ? for help')
      else:
        await self._helper.run(v)
    elif op == '=':
      if val is not None:
        self._aliases[n] = val
      else:
        del self._aliases[n]

  async def _alias2(self, line: str) -> None:
    lex = shlex(line, posix=True, punctuation_chars=';=')
    lex.wordchars += '@:,!$'

    args = deque(lex)
    args.popleft()

    newcmd = args.popleft()
    if not re.fullmatch('-?[a-zA-Z0-9_]+', newcmd):
      ui.fatal(f'invalid macro name: {newcmd}')

    argn = 0
    while args:
      t = args.popleft()
      if re.fullmatch(';+', t) or t == ')':
        break
      elif not re.fullmatch('[a-zA-Z0-9_]+', t):
        ui.fatal(f'invalid arg name: {t}')
      else:
        argn += 1

    body: deque[str] = deque()
    while args:
      t = args.popleft()
      if t == ')':
        break
      else:
        body.append(t)

    if body:
      if newcmd.startswith('-'):
        ui.fatal(f'invalid macro name: {newcmd}')
      for t in body:
        m = re.search(r'\$([0-9]+)', t)
        if m:
          nr = int(m.group(1))
          if not nr < argn:
            ui.fatal('arg index out of range: {} (macro takes {} args)'.format(m.group(0), argn))
    else:
      if not newcmd.startswith('-'):
        ui.fatal('invalid macro: no body found')
      else:
        cmd = newcmd[1:]
        try:
          del self._macros[cmd]
        except KeyError:
          ui.fatal(f'macro not found: {cmd}')
        return

    self._macros[newcmd] = argn, line, body

  async def _alias2_call(self, line: str) -> None:
    content = re.match(r'\.\((.+)\)', line)
    assert content is not None
    lex = shlex(content.group(1), posix=True, punctuation_chars=';=')
    lex.wordchars += '@:,!$'

    tokens = deque(lex)
    cmd = tokens.popleft()

    argn, _, body = self._macros[cmd]
    args = []

    for _ in range(argn):
      try:
        t = tokens.popleft()
      except IndexError:
        ui.fatal('not enough arg (requires {})'.format(argn))
      else:
        args.append(t)

    if tokens:
      ui.warn('igonring extra {} args'.format(len(tokens)))

    asl: deque[str] = deque()
    for t in body:
      asl.append(re.sub(r'\$([0-9]+)', lambda m: args[int(m.group(1))], t))
    await self._helper.run_cmd(asl, None)



from __future__ import annotations
from typing import TYPE_CHECKING
from contextlib import asynccontextmanager
from functools import cache
from os import read, write, close

if TYPE_CHECKING:
  from aiohttp import ClientSession
  from typing import Protocol, Optional, AsyncIterator, AsyncContextManager, Tuple
  from typing_extensions import Self

  class SwiftDemanglerImpl(Protocol):
    def scoped(self) -> AsyncContextManager[Self]: ...
    async def resolve(self, q: str) -> str: ...

class SwiftDemangler:
  @classmethod
  def get(cls, simplify: bool = False) -> SwiftDemanglerImpl:
    if _find_swift_demangler():
      return _Local(simplify)
    else:
      return _Remote(simplify)

class _Local:
  _f: Optional[Tuple[int, int]]

  def __init__(self, simplify: bool = False):
    from asyncio import Lock, get_running_loop
    from threading import Event
    self._l = get_running_loop()
    self._f = None
    self._w = Event()
    self._s = Lock()
    self._simp = simplify

  @asynccontextmanager
  async def scoped(self) -> AsyncIterator[Self]:
    try:
      yield self
    finally:
      if self._f:
        for n in self._f:
          close(n)

  async def _boot(self) -> None:
    def _do() -> None:
      from subprocess import Popen
      from os import openpty
      assert self._f is None
      i0, i1 = openpty()
      o0, o1 = openpty()
      with Popen(self._get_cmdline(), shell=True, bufsize=0, stdin=i1, stdout=o1):
        self._f = (i0, o0)
        self._w.set()
      self._w.clear()
      self._f = None

    if not self._w.is_set():
      self._l.run_in_executor(None, _do)
      self._w.wait()

  async def resolve(self, q: str) -> str:
    async with self._s:
      await self._boot()
      assert self._f is not None
      write(self._f[0], q.encode() + b'\n')
      o = bytearray()
      while not o.endswith(b'\n'):
        o = o + read(self._f[1], 1024)
      return o.rstrip().decode()

  def _get_cmdline(self) -> str:
    path = _find_swift_demangler()
    assert path
    return '{}{}'.format(
      path,
      ' -simplified' if self._simp else ''
    )

class _Remote:
  _sess: Optional[ClientSession] = None

  def __init__(self, simplify: bool = False):
    from trueseeing.core.env import get_swift_demangler_url
    self._url = '{base}{prefix}'.format(base=get_swift_demangler_url(), prefix='/s/' if simplify else '/')

  @asynccontextmanager
  async def scoped(self) -> AsyncIterator[Self]:
    from aiohttp import ClientSession, ClientConnectionError
    async with ClientSession() as sess:
      try:
        async with sess.get(self._url + 'a'):
          pass
        self._sess = sess
      except ClientConnectionError:
        from trueseeing.core.ui import ui
        ui.warn('swift demangler is not available (try booting the demangler container and attaching it to the network)')
        self._sess = None
      yield self
    self._sess = None

  async def resolve(self, q: str) -> str:
    if self._sess:
      async with self._sess.get(f'{self._url}{q}') as resp:
        return (await resp.json())['to'] # type:ignore[no-any-return]
    else:
      return q

@cache
def _find_swift_demangler() -> Optional[str]:
  from subprocess import run, CalledProcessError
  try:
    if run('which swift-demangle', shell=True, capture_output=True).stdout:
      return 'swift-demangle'
    elif run('which swift', shell=True, capture_output=True).stdout:
      return 'swift demangle'
    else:
      return None
  except CalledProcessError:
    return None



from __future__ import annotations
import zstandard as zstd

def zd(item: bytes) -> bytes:
  return zstd.ZstdDecompressor().decompress(item)

def ze(item: bytes) -> bytes:
  return zstd.ZstdCompressor(threads=-1).compress(item)



from __future__ import annotations
from typing import TYPE_CHECKING
import io
import os
import re

from trueseeing.core.android.model import InvocationPattern
from trueseeing.core.android.model import SignatureMixin
from trueseeing.core.android.analyze.flow import DataFlow

if TYPE_CHECKING:
  from typing import Iterable, Optional, List, Dict, Any, Set
  from trueseeing.api import Signature, SignatureHelper, SignatureMap
  from trueseeing.core.model.issue import IssueConfidence

# XXX huge duplication
class TopLevelSuffixes:
  def __init__(self) -> None:
    from importlib.resources import files
    with (files('trueseeing')/'libs'/'tlds.txt').open('r', encoding='utf-8') as f:
      self._re_tlds = re.compile('^(?:{})$'.format('|'.join(re.escape(l.strip()) for l in f if l and not l.startswith('#'))), flags=re.IGNORECASE)

  def looks_public(self, names: List[str]) -> bool:
    if names:
      gtld = names[0]
      return (gtld == 'android') or bool(self._re_tlds.search(gtld))
    else:
      return False

class PublicSuffixes:
  _suffixes: Set[str] = set()

  def __init__(self) -> None:
    from importlib.resources import files
    with (files('trueseeing')/'libs'/'public_suffix_list.dat').open('r', encoding='utf-8') as f:
      self._suffixes.update((l for l in f if l and not l.startswith('//')))

  def looks_public(self, names: List[str]) -> bool:
    suf = '.'.join(reversed(names))
    return suf in self._suffixes

class LibraryDetector(SignatureMixin):
  _id = 'detect-library'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'

  _suffixes_top: TopLevelSuffixes = TopLevelSuffixes()
  _suffixes_public: PublicSuffixes = PublicSuffixes()

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return LibraryDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id: dict(e=self.detect, d='Detects libraries')}

  @classmethod
  def _package_name_of(cls, path: str) -> str:
    return os.path.dirname(path).replace('/', '.')

  @classmethod
  def _package_family_of(cls, p: str) -> Optional[str]:
    f = {
      r'javax\..*': None,
      r'(android\.support\.v[0-9]+)\..*': r'\1',
      r'(com\.google\.android\.gms)\..*': r'\1',
      r'(.*?)\.internal(?:\..*)?$': r'\1',
      r'(.*?)(?:\.[a-z]{,4})+$': r'\1',
      r'([a-z0-9_]{5,}(?:\.[a-z0-9_]{2,})+?)\..*': r'\1',
    }
    for k, v in f.items():
      if re.match(k, p):
        if v is not None:
          return re.sub(k, v, p)
        else:
          return None

    return p

  @classmethod
  def _shared_package_of(cls, c1: str, c2: str) -> Optional[List[str]]:
    o = []
    try:
      for a,b in zip(c1.split('.'), c2.split('.')):
        if a == b:
          o.append(a)
        else:
          break
    finally:
      return o

  @classmethod
  def _is_kind_of(cls, c1: str, c2: str) -> bool:
    shared = cls._shared_package_of(c1, c2)
    if shared is None:
      return False
    elif not len(shared):
      return False
    elif len(shared) == 1:
      return not cls._suffixes_top.looks_public(shared)
    elif len(shared) > 2:
      return True
    else:
      return not cls._suffixes_public.looks_public(shared)

  async def detect(self) -> None:
    context = self._get_context()
    q = context.store().query()
    package = context.parsed_manifest().xpath('/manifest/@package', namespaces=dict(android='http://schemas.android.com/apk/res/android'))[0]

    packages: Dict[str, List[str]] = dict()
    for fn in (context.source_name_of_disassembled_class(r) for r in context.disassembled_classes()):
      # XXX exclude packages
      family = self._package_family_of(self._package_name_of(fn))
      if family is not None:
        try:
          packages[family].append(fn)
        except KeyError:
          packages[family] = [fn]
        else:
          pass
    packages = {k:v for k,v in packages.items() if not self._is_kind_of(k, package) and re.search(r'\.[a-zA-Z0-9]{4,}(?:\.|$)', k)}

    for p in sorted(packages.keys()):
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cvss=self._cvss,
        title='detected library',
        info0=f'{p} (score: {len(packages[p])})',
      ))

    for p in reversed(sorted(packages.keys(), key=len)):
      for k in q.consts_in_package(p, InvocationPattern('const-string', r'[0-9]+\.[0-9]+|(19|20)[0-9]{2}[ /-]')):
        ver = self._an.get_param(k, 1).v
        if not re.search(r'^/|:[0-9]+|\\|://', ver):
          comps = ver.split('.')
          if len(comps) > 4:
            continue
          if ' and ' not in ver and re.match('^[0-9]|^v[0-9]', ver):
            if len(comps) < 4 or self._comp4_looks_like_version(comps):
              self._helper.raise_issue(self._helper.build_issue(
                sigid=self._id,
                cvss=self._cvss,
                title='detected library version',
                info0=f'{ver} ({p})',
              ))
            else:
              self._helper.raise_issue(self._helper.build_issue(
                sigid=self._id,
                cfd='tentative',
                cvss=self._cvss,
                title='potential library version',
                info0=f'{ver} ({p})',
              ))
          else:
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cfd='tentative',
              cvss=self._cvss,
              title='potential version/dated reference in library',
              info0=f'{ver} ({p})',
            ))

    for fn, blob in q.file_enum('root/assets/%.js'):
      f = io.StringIO(blob.decode('utf-8', errors='ignore'))
      for l in f:
        for m in re.finditer(r'[0-9]+\.[0-9]+|(19|20)[0-9]{2}[ /-]', l):
          ver = l
          if not re.search(r'^/|:[0-9]+|\\|://|[\x00-\x1f]', ver):
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cfd='tentative',
              cvss=self._cvss,
              title='potential version/dated reference in library',
              info0='{match} ({rfn})'.format(rfn=fn, match=ver),
            ))

  def _comp4_looks_like_version(self, xs: List[str]) -> bool:
    if xs[0].lower().startswith('v'):
      return True
    ints = []
    for c in xs:
      try:
        ints.append(int(c))
      except ValueError:
        return True
    if 0 in ints or any([(t > 255) for t in ints]):
      return True
    elif all([(t < 100) for t in ints[:2]]):
      return True
    else:
      return False

class ProGuardDetector(SignatureMixin):
  _id = 'detect-obfuscator'
  _cvss_true = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _cvss_false = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:L/I:N/A:N/'

  _whitelist = ['R']

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ProGuardDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects obfuscators')}

  @classmethod
  def _class_name_of(self, path: str) -> str:
    return path.replace('.smali', '').replace('/', '.')

  async def detect(self) -> None:
    context = self._get_context()
    for c in (self._class_name_of(context.source_name_of_disassembled_class(r)) for r in context.disassembled_classes()):
      m = re.search(r'(?:^|\.)(.)$', c)
      if m and m.group(1) not in self._whitelist:
        self._helper.raise_issue(self._helper.build_issue(sigid=self._id, cfd='certain', cvss=self._cvss_true, title='detected obfuscator', info0='ProGuard'))
        break
    else:
      self._helper.raise_issue(self._helper.build_issue(sigid=self._id, cvss=self._cvss_false, title='lack of obfuscation'))

class UrlLikeDetector(SignatureMixin):
  _id = 'detect-url'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'

  _re_tlds: Optional[re.Pattern[str]] = None

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return UrlLikeDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {'detect-url':dict(e=self.detect, d='Detects URL-like strings')}

  def _analyzed(self, x: str, qn: Optional[str] = None) -> Iterable[Dict[str, Any]]:
    assert self._re_tlds is not None
    if '://' in x:
      if re.search(r'//[{}$%a-zA-Z0-9_:-]+?@[{}$%a-zA-Z0-9_-]', x):
        yield dict(
          type_='authenticated URL', value=re.findall(r'\S+://\S+', x), cfd='firm',
          cvss='CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:N/',
        )
      else:
        yield dict(type_='URL', value=re.findall(r'\S+://\S+', x), cfd='firm')
    elif re.search(r'^/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+', x):
      yield dict(type_='path component', value=re.findall(r'^/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+', x), cfd='firm')
    elif re.search(r'^[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)+(:[0-9]+)?$', x):
      m = re.search(r'^([^:/]+)', x)
      if m:
        hostlike = m.group(1)
        components = hostlike.split('.')
        if len(components) == 4 and all(re.match(r'^\d+$', c) for c in components) and all(int(c) < 256 for c in components):
          if re.match(r'1\.[239]\.|2\.(1|5|160)\.|3\.1\.', hostlike) and qn is not None and re.search(r'asn1|ASN1|x509|X509|KeyUsage', qn):
            pass
          else:
            yield dict(type_='possible IPv4 address', value=[hostlike], cfd='firm')
        elif self._re_tlds.search(components[-1]):
          if not re.search(r'^android\.(intent|media)\.|^os\.name$|^java\.vm\.name|^[A-Z]+.*\.(java|EC|name|secure)$|^Dispatchers\.IO', hostlike):
            confidence: IssueConfidence = 'firm'
            if re.search(r'^(java|os|kotlin|google\.[a-z]+?)\.|^(com|cn)\.google|\.(date|time|host|help|prof|name|top|link|id|icu|app|[a-z]{5,})$', hostlike, flags=re.IGNORECASE):
              confidence = 'tentative'
            yield dict(type_='possible FQDN', value=[hostlike], cfd=confidence)

  async def detect(self) -> None:
    from importlib.resources import files
    with (files('trueseeing')/'libs'/'tlds.txt').open('r', encoding='utf-8') as f:
      self._re_tlds = re.compile('^(?:{})$'.format('|'.join(re.escape(l.strip()) for l in f if l and not l.startswith('#'))), flags=re.IGNORECASE)

    context = self._get_context()
    q = context.store().query()
    for cl in q.consts(InvocationPattern('const-string', r'://|^/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+|^[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)+(:[0-9]+)?$')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      for match in self._analyzed(self._an.get_param(cl, 1).v, qn):
        for v in match['value']:
          self._helper.raise_issue(self._helper.build_issue(sigid=self._id, cfd=match['cfd'], cvss=match.get('cvss', self._cvss), title=f'detected {match["type_"]}', info0=v, aff0=qn))
    for name, val in context.string_resources():
      for match in self._analyzed(val):
        for v in match['value']:
          self._helper.raise_issue(self._helper.build_issue(sigid=self._id, cfd=match['cfd'], cvss=match.get('cvss', self._cvss), title=f'detected {match["type_"]}', info0=v, aff0=f'R.string.{name}'))

class NativeMethodDetector(SignatureMixin):
  _id = 'detect-native-method'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary = 'Natively defined methods'
  _synopsis = "The application uses JNI."
  _detailed_description = None

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return NativeMethodDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects natively defined methods')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for op in q.methods_with_modifier('native'):
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cvss=self._cvss,
        title=self._summary,
        summary=self._synopsis,
        aff0=q.qualname_of(op.addr)
      ))

class NativeArchDetector(SignatureMixin):
  _id = 'detect-native-arch'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary = 'Supported architectures'
  _synopsis = "The application has native codes for some architectures."

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return NativeArchDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects supported architectures')}

  async def detect(self) -> None:
    for d in self._get_context().store().query().file_find('root/lib/%'):
      if re.search(r'arm|x86|mips', d):
        arch = d.split('/')[2]
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cvss=self._cvss,
          title=self._summary,
          info0=arch,
          info1=os.path.basename(d),
          summary=self._synopsis,
        ))

class ReflectionDetector(SignatureMixin):
  _id = 'detect-reflection'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary0 = 'Use of reflection'
  _synopsis0 = "The application makes use of Java reflection APIs."
  _summary1 = 'Classloader reference'
  _synopsis1 = "The application makes use of classloaders."

  _blacklist_caller = [
    r'Lkotlin/',
    r'Lkotlinx/',
  ]

  _blacklist_meth = [
    r'Ljava/lang/Object;->getClass\(\)',
    r'Ljava/lang/Class;->getClassLoader',
    r'Ljava/lang/Class;->get((Simple|Canonical)?Name|ComponentType|Annotation)',
    r'Ljava/lang/Class;->.*?\(\)',
    r'Ljava/lang/reflect/Method;->.*?\(\)',
    r'Ljava/lang/reflect/Field;->.*?\(\)',
  ]

  _blacklist_val = [
    r'^-?0x[0-9a-f]+$'
  ]

  _masker = '(unknown)'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ReflectionDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects reflections')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-', '^Ljavax?.*/(Class|Method|Field);->|^Ljava/lang/[A-Za-z]*?ClassLoader;->')):
      qn = q.qualname_of(cl.addr)
      if qn:
        if context.is_qualname_excluded(qn):
          continue
        if any(re.match(x, qn) for x in self._blacklist_caller):
          continue
      ct = q.method_call_target_of(cl.addr)
      if ct is None:
        continue
      if any(re.match(x, ct) for x in self._blacklist_meth):
        continue
      if 'ClassLoader;->' in ct:
        try:
          xs = DataFlow(q).solved_possible_constant_data_in_invocation(cl, 0)
          if not xs:
            xs = {self._masker}
          for x in xs:
            if any(re.match(p, x) for p in self._blacklist_val):
              x = self._masker
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss,
              title=self._summary1,
              info0=ct,
              info1=x,
              aff0=qn,
              summary=self._synopsis1,
              desc=self._synopsis1,
            ))
        except IndexError:
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss,
            title=self._summary1,
            info0=ct,
            aff0=qn,
            summary=self._synopsis1,
            desc=self._synopsis1,
          ))
      else:
        try:
          xs = DataFlow(q).solved_possible_constant_data_in_invocation(cl, 0)
          if not xs:
            xs = {self._masker}
          for x in xs:
            if any(re.match(p, x) for p in self._blacklist_val):
              x = self._masker
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss,
              title=self._summary0,
              info0=ct,
              info1=x,
              aff0=qn,
              summary=self._synopsis0,
              desc=self._synopsis0,
            ))
        except IndexError:
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss,
            title=self._summary0,
            info0=ct,
            aff0=qn,
            summary=self._synopsis0,
            desc=self._synopsis0,
          ))



from __future__ import annotations
from typing import TYPE_CHECKING

import itertools
import io
import re
import os

from trueseeing.core.android.model import InvocationPattern
from trueseeing.core.android.model import SignatureMixin
from trueseeing.core.android.analyze.flow import DataFlow
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Iterable, Optional, Set, Tuple, Any, TypeVar, Dict
  from trueseeing.api import Signature, SignatureHelper, SignatureMap
  from trueseeing.core.model.issue import IssueConfidence
  T = TypeVar('T')

class SecurityFilePermissionDetector(SignatureMixin):
  _id = 'security-file-permission'
  _cvss = 'CVSS:3.0/AV:L/AC:L/PR:N/UI:N/S:C/C:L/I:L/A:L/'
  _summary = 'insecure file permission'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecurityFilePermissionDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects insecure file creation')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-virtual', r'Landroid/content/Context;->openFileOutput\(Ljava/lang/String;I\)')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        target_val = int(DataFlow(q).solved_constant_data_in_invocation(cl, 1), 16)
        if target_val & 3:
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cfd='certain',
            cvss=self._cvss,
            title=self._summary,
            info0={1: 'MODE_WORLD_READABLE', 2: 'MODE_WORLD_WRITEABLE'}[target_val],
            aff0=q.qualname_of(cl.addr)
          ))
      except (DataFlow.NoSuchValueError):
        pass


class SecurityTlsInterceptionDetector(SignatureMixin):
  _id = 'security-tls-interception'
  _cvss = 'CVSS:3.0/AV:N/AC:H/PR:H/UI:R/S:C/C:L/I:L/A:L/'
  _cvss_info = 'CVSS:3.0/AV:N/AC:L/PR:L/UI:N/S:U/C:N/I:N/A:N/'
  _summary = 'insecure TLS connection'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecurityTlsInterceptionDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects certificate (non-)pinning')}

  async def detect(self) -> None:
    context = self._get_context()
    pin_nsc = False
    if context.get_min_sdk_version() > 23:
      if not context.parsed_manifest().xpath('//application[@android:debuggable="true"]', namespaces=dict(android='http://schemas.android.com/apk/res/android')):
        pin_nsc = True

    for fn, xp in context.xml_resources():
      if 'network-security-config' in xp.tag.lower():
        for e in xp.xpath('.//certificates'):
          if e.attrib.get('src') == 'user':
            pin_nsc = False
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss,
              title=self._summary,
              info0='user-trusting network security config detected'
            ))
          for pin in e.xpath('.//pins'):
            algo: str
            dig: str
            algo, dig = pin.attrib('digest', '(unknown)'), pin.text
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss_info,
              title='explicit ceritifcate pinning',
              info0=f'{algo}:{dig}',
            ))
    if not pin_nsc:
      if not self._do_detect_plain_pins_x509():
        if not self._do_detect_plain_pins_hostnameverifier():
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss,
            title=self._summary,
            info0='no pinning detected'
          ))

  def _do_detect_plain_pins_x509(self) -> Set[str]:
    context = self._get_context()
    pins: Set[str] = set()
    store = context.store()
    q = store.query()
    for addr in q.methods_in_class('checkServerTrusted', 'X509TrustManager'):
      if any(q.matches_in_method(addr, InvocationPattern('verify', ''))):
        classname = q.class_name_of(addr)
        if classname:
          pins.add(classname)
      if any(q.matches_in_method(addr, InvocationPattern('throw', ''))):
        classname = q.class_name_of(addr)
        if classname:
          pins.add(classname)

    if pins:
      # XXX crude detection
      custom_sslcontext_detected = False
      for cl in q.invocations(InvocationPattern('invoke-virtual', 'Ljavax/net/ssl/SSLContext;->init')):
        custom_sslcontext_detected = True
        pins = DataFlow(q).solved_typeset_in_invocation(cl, 1) & pins

      if not custom_sslcontext_detected:
        return set()
      else:
        return pins
    else:
      return pins

  def _do_detect_plain_pins_hostnameverifier(self) -> Set[str]:
    context = self._get_context()
    pins: Set[str] = set()
    q = context.store().query()
    for addr in itertools.chain(q.methods_in_class('verify(Ljava/lang/String;Ljavax/net/ssl/SSLSession;)Z', 'HostnameVerifier')):
      if any(q.matches_in_method(addr, InvocationPattern('invoke', 'contains|equals|verify|Ljavax/net/ssl/SSLSession;->getPeerCertificates'))):
        classname = q.class_name_of(addr)
        if classname:
          pins.add(classname)
    return pins


class LayoutSizeGuesser:
  _xmlns_android = '{http://schemas.android.com/apk/res/android}'
  _table = {'small':(320.0, 426.0), 'normal':(320.0, 470.0), 'large':(480.0, 640.0), 'xlarge':(720.0, 960.0)}

  def guessed_size(self, t: Any, path: str) -> float:
    dps = self._dps_from_modifiers(self._modifiers_in(path))
    for e in self._self_and_containers_of(t):
      try:
        width, height = self._width_of(e), self._height_of(e)
      except KeyError:
        try:
          ui.warn('layout_guesser: guessed_size: ignoring improper webview declaration ({0})'.format(e.attrib[f'{self._xmlns_android}id']))
          return 0.0
        except KeyError:
          ui.warn('layout_guesser: guessed_size: ignoring improper webview declaration')
          return 0.0
      else:
        if any(self._is_bound(x) for x in (width, height)):
          return self._guessed_dp(width, dps[0]) * self._guessed_dp(height, dps[1])
    else:
      return 1.0

  @classmethod
  def _dps_from_modifiers(cls, mods: Set[str]) -> Tuple[float, float]:
    try:
      x, y = cls._table[list(mods & cls._table.keys())[0]]
    except (IndexError, KeyError):
      x, y = cls._table['large']
    if 'land' in mods:
      return (y, x)
    else:
      return (x, y)

  @classmethod
  def _width_of(cls, e: Any) -> str:
    return e.attrib[f'{cls._xmlns_android}layout_width'] # type:ignore[no-any-return]

  @classmethod
  def _height_of(cls, e: Any) -> str:
    return e.attrib[f'{cls._xmlns_android}layout_height'] # type:ignore[no-any-return]

  @classmethod
  def _is_bound(cls, x: str) -> bool:
    return x not in ('fill_parent', 'match_parent', 'wrap_content')

  @classmethod
  def _guessed_dp(cls, x: str, dp: float) -> float:
    if cls._is_bound(x):
      try:
        return float(re.sub(r'di?p$', '', x)) / float(dp)
      except ValueError:
        try:
          ui.debug("layout_guesser: guessed_size: guessed_dp: warning: ignoring non-dp suffix ({!s})".format(x))
          return float(re.sub(r'[^0-9-]', '', x)) / float(dp)
        except ValueError:
          ui.debug("layout_guesser: guessed_size: guessed_dp: warning: ignoring unknown dimension")
          return 0.0
    else:
      return dp

  @classmethod
  def _self_and_containers_of(cls, e: Any) -> Iterable[Any]:
    yield e
    e = e.getparent()
    if e is not None:
      cls._self_and_containers_of(e)

  @classmethod
  def _modifiers_in(cls, path: str) -> Set[str]:
    return [set(c.split('-')) for c in path.split(os.sep) if 'layout' in c][0]


class SecurityTamperableWebViewDetector(SignatureMixin):
  _id = 'security-tamperable-webview'
  description = 'Detects tamperable WebView'
  _summary1 = 'tamperable webview'
  _summary2 = 'tamperable webview with URL'
  _cvss1 = 'CVSS:3.0/AV:N/AC:H/PR:N/UI:R/S:U/C:N/I:L/A:L/'
  _cvss2 = 'CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:U/C:N/I:L/A:L/'

  _xmlns_android = '{http://schemas.android.com/apk/res/android}'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecurityTamperableWebViewDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects tamperable WebView')}

  async def detect(self) -> None:
    import lxml.etree as ET
    from functools import reduce

    context = self._get_context()
    store = context.store()
    q = store.query()
    targets = {'WebView','XWalkView','GeckoView'}

    more = True
    while more:
      more = False
      for cl in q.related_classes('|'.join(targets)):
        name = q.class_name_of(cl)
        if name is not None and name not in targets:
          targets.add(name)
          more = True

    for fn, blob in q.file_enum('%/res/%layout%.xml'):
      r = ET.fromstring(blob, parser=ET.XMLParser(recover=True))
      for t in reduce(lambda x,y: x+y, (r.xpath('//{}'.format(context.class_name_of_dalvik_class_type(c).replace('$', '_'))) for c in targets)):
        size = LayoutSizeGuesser().guessed_size(t, fn)
        if size > 0.5:
          try:
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cfd='tentative',
              cvss=self._cvss1,
              title=self._summary1,
              info0='{0} (score: {1:.02f})'.format(t.attrib[f'{self._xmlns_android}id'], size),
              aff0=context.source_name_of_disassembled_resource(fn)
            ))
          except KeyError as e:
            ui.warn(f'SecurityTamperableWebViewDetector.do_detect: missing key {e}')

    # XXX: crude detection
    for op in q.invocations(InvocationPattern('invoke-', ';->loadUrl')):
      qn = q.qualname_of(op.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        v = DataFlow(q).solved_constant_data_in_invocation(op, 0)
        if v.startswith('http://'):
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss2,
            title=self._summary2,
            info0=v,
            aff0=q.qualname_of(op.addr)
          ))
      except DataFlow.NoSuchValueError:
        pass


class SecurityInsecureWebViewDetector(SignatureMixin):
  _id = 'security-insecure-webview'
  _cvss = 'CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:C/C:L/I:L/A:L/'
  _cvss2 = 'CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:L/A:L/'
  _cvss2b = 'CVSS:3.0/AV:N/AC:H/PR:N/UI:N/S:U/C:L/I:N/A:N/'
  _cvss3 = 'CVSS:3.0/AV:N/AC:H/PR:N/UI:N/S:U/C:N/I:L/A:N/'
  _cvss4 = 'CVSS:3.0/AV:N/AC:H/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary1 = 'insecure Javascript interface'
  _summary2 = 'insecure mixed content mode'
  _summary2b = 'potentially insecure mixed content mode'
  _summary3 = 'insecure CSP'
  _summary4 = 'detected CSP'

  _xmlns_android = '{http://schemas.android.com/apk/res/android}'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecurityInsecureWebViewDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect,d='Detects insecure WebView')}

  # FIXME: Come up with something more right
  @classmethod
  def _first(cls, xs: Iterable[T], default: Optional[T] = None) -> Optional[T]:
    try:
      return list(itertools.islice(xs, 1))[0]
    except IndexError:
      return default

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    query = store.query()

    targets = set()
    seeds = {'WebView','XWalkView','GeckoView'}

    more = True
    while more:
      more = False
      for addr in store.query().related_classes('|'.join(seeds)):
        name = store.query().class_name_of(addr)
        if name not in targets:
          targets.add(name)
          more = True
    for seed in seeds:
      targets.add(f'L.*{seed};')

    # XXX: Crude detection
    # https://developer.android.com/reference/android/webkit/WebView.html#addJavascriptInterface(java.lang.Object,%2520java.lang.String)
    if context.get_min_sdk_version() <= 16:
      for p in query.invocations(InvocationPattern('invoke-virtual', 'Landroid/webkit/WebSettings;->setJavaScriptEnabled')):
        qn = query.qualname_of(p.addr)
        if context.is_qualname_excluded(qn):
          continue
        try:
          if DataFlow(query).solved_constant_data_in_invocation(p, 0):
            for target in targets:
              for q in query.invocations_in_class(p.addr, InvocationPattern('invoke-virtual', f'{target}->addJavascriptInterface')):
                try:
                  if DataFlow(query).solved_constant_data_in_invocation(q, 0):
                    self._helper.raise_issue(self._helper.build_issue(
                      sigid=self._id,
                      cvss=self._cvss,
                      title=self._summary1,
                      aff0=query.qualname_of(q.addr)
                    ))
                except (DataFlow.NoSuchValueError):
                  self._helper.raise_issue(self._helper.build_issue(
                    sigid=self._id,
                    cfd='tentative',
                    cvss=self._cvss,
                    title=self._summary1,
                    aff0=query.qualname_of(q.addr)
                  ))
        except (DataFlow.NoSuchValueError):
          pass

    # https://developer.android.com/reference/android/webkit/WebSettings#setMixedContentMode(int)
    if context.get_min_sdk_version() >= 21:
      for q in query.invocations(InvocationPattern('invoke-virtual', 'Landroid/webkit/WebSettings;->setMixedContentMode')):
        qn = query.qualname_of(q.addr)
        if context.is_qualname_excluded(qn):
          continue
        try:
          val = int(DataFlow(query).solved_constant_data_in_invocation(q, 0), 16)
          if val == 0:
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss2,
              title=self._summary2,
              info0='MIXED_CONTENT_ALWAYS_ALLOW',
              aff0=query.qualname_of(q.addr)))
          elif val == 2:
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cvss=self._cvss2b,
              title=self._summary2b,
              info0='MIXED_CONTENT_COMPATIBILITY_MODE',
              aff0=query.qualname_of(q.addr)))
        except (DataFlow.NoSuchValueError):
          pass
    else:
      for target in targets:
        for q in query.invocations(InvocationPattern('invoke-virtual', f'{target}->loadUrl')):
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss,
            title=self._summary2,
            info0='mixed mode always enabled in API < 21',
            aff0=query.qualname_of(q.addr)
          ))

    for op in query.invocations(InvocationPattern('invoke-', ';->loadUrl')):
      qn = query.qualname_of(op.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        v = DataFlow(query).solved_constant_data_in_invocation(op, 0)
        if v.startswith('file:///android_asset/'):
          path = v.replace('file:///android_asset/', 'assets/')
          blob = query.file_get(f'root/{path}')
          if blob is not None:
            content = blob.decode('utf-8', errors='ignore')
            m = re.search('<meta .*Content-Security-Policy.*content="(.*)?">', content, flags=re.IGNORECASE)
            csp: Optional[str] = None if m is None else m.group(1)
            if csp is None or any([(x in csp.lower()) for x in ('unsafe', 'http:')]):
              self._helper.raise_issue(self._helper.build_issue(
                sigid=self._id,
                cvss=self._cvss3,
                title=self._summary3,
                info0=path,
                info1='default' if csp is None else csp,
                aff0=query.qualname_of(op.addr)
              ))
            else:
              self._helper.raise_issue(self._helper.build_issue(
                sigid=self._id,
                cvss=self._cvss4,
                title=self._summary4,
                info0=path,
                info1=csp,
                aff0=query.qualname_of(op.addr)
              ))
      except DataFlow.NoSuchValueError:
        pass

class FormatStringDetector(SignatureMixin):
  _id = 'security-format-string'
  _summary = 'detected format string'
  _cvss = 'CVSS:3.0/AV:P/AC:H/PR:N/UI:N/S:U/C:N/I:N/A:N/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return FormatStringDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects format string usages')}

  def _analyzed(self, x: str) -> Iterable[Dict[str, Any]]:
    if re.search(r'%s', x):
      if re.search(r'(://|[<>/&?])', x):
        yield dict(cfd='firm', value=x)

  async def detect(self) -> None:
    context = self._get_context()
    q = context.store().query()
    for cl in q.consts(InvocationPattern('const-string', r'%s')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      for t in self._analyzed(self._an.get_param(cl, 1).v):
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd=t['cfd'],
          cvss=self._cvss,
          title=self._summary,
          info0=t['value'],
          aff0=q.qualname_of(cl.addr)
        ))
    for name, val in context.string_resources():
      for t in self._analyzed(val):
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd=t['cfd'],
          cvss=self._cvss,
          title=self._summary,
          info0=t['value'],
          aff0=f'R.string.{name}'
        ))

class LogDetector(SignatureMixin):
  _id = 'security-log'
  _summary = 'detected logging'
  _cvss = 'CVSS:3.0/AV:P/AC:H/PR:N/UI:N/S:U/C:L/I:N/A:N/'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return LogDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects logging activities')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-', r'L.*->([dwie]|debug|error|exception|warning|info|notice|wtf)\(Ljava/lang/String;Ljava/lang/String;.*?Ljava/lang/(Throwable|.*?Exception);|L.*;->print(ln)?\(Ljava/lang/String;|LException;->printStackTrace\(')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue

      func = self._an.get_param(cl, 1).v
      if 'print' not in func:
        try:
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cfd='tentative',
            cvss=self._cvss,
            title=self._summary,
            info0=func,
            info1=DataFlow(q).solved_constant_data_in_invocation(cl, 1),
            aff0=q.qualname_of(cl.addr)
          ))
        except (DataFlow.NoSuchValueError):
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cfd='tentative',
            cvss=self._cvss,
            title=self._summary,
            info0=func,
            aff0=q.qualname_of(cl.addr)
          ))
      elif 'Exception;->' not in func:
        try:
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cfd='tentative',
            cvss=self._cvss,
            title=self._summary,
            info0=func,
            info1=DataFlow(q).solved_constant_data_in_invocation(cl, 0),
            aff0=q.qualname_of(cl.addr)
          ))
        except (DataFlow.NoSuchValueError):
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cfd='tentative',
            cvss=self._cvss,
            title=self._summary,
            info0=func,
            aff0=q.qualname_of(cl.addr)
          ))
      else:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='tentative',
          cvss=self._cvss,
          title=self._summary,
          info0=func,
          aff0=q.qualname_of(cl.addr)
        ))

class ADBProbeDetector(SignatureMixin):
  _id = 'security-adb-detect'
  _cvss = 'CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary = 'USB debugging detection'
  _synopsis = 'The application is probing for USB debugging (adbd.)'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ADBProbeDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects probe of adbd status.')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-', r'^Landroid/provider/Settings\$(Global|Secure);->getInt\(')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      for found in DataFlow(q).solved_possible_constant_data_in_invocation(cl, 1):
        if found == 'adb_enabled':
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss,
            title=self._summary,
            aff0=q.qualname_of(cl.addr),
            summary=self._synopsis,
          ))

class ClientXSSJQDetector(SignatureMixin):
  _id = 'security-cxss-jq'
  _cvss = 'CVSS:3.0/AV:N/AC:H/PR:N/UI:R/S:U/C:L/I:L/A:N/'
  _summary = 'Potential client-side XSS (JQuery)'
  _synopsis = "The application pours literal HTML in JQuery context."

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return ClientXSSJQDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects potential client-side XSS vector in JQuery-based apps')}

  async def detect(self) -> None:
    context = self._get_context()
    for fn, blob in context.store().query().file_enum(pat='root/assets/%.js'):
      f = io.StringIO(blob.decode('utf-8', errors='ignore'))
      for l in f:
        for m in re.finditer(r'\.html\(', l):
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cvss=self._cvss,
            title=self._summary,
            info0='{match} ({rfn})'.format(rfn=fn, match=l),
            summary=self._synopsis,
          ))

class SecurityFileWriteDetector(SignatureMixin):
  _id = 'security-file-write'
  _cvss1 = 'CVSS:3.0/AV:L/AC:H/PR:N/UI:N/S:U/C:L/I:N/A:N/'
  _summary1 = 'detected potential logging into file'
  _synopsis1 = 'The application is potentially logging into file.'
  _cvss2 = 'CVSS:3.0/AV:L/AC:H/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary2 = 'open files for writing'
  _synopsis2 = 'The application opens files for writing.'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecurityFileWriteDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects file creation')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-virtual', r'Landroid/content/Context;->openFileOutput\(Ljava/lang/String;I\)')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        target_val = DataFlow(q).solved_constant_data_in_invocation(cl, 0)
      except DataFlow.NoSuchValueError:
        target_val = '(unknown name)'

      if re.search(r'debug|log|info|report|screen|err|tomb|drop', target_val):
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss1,
          title=self._summary1,
          summary=self._synopsis1,
          info0=target_val,
          aff0=q.qualname_of(cl.addr)
        ))
      else:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss2,
          title=self._summary2,
          summary=self._synopsis2,
          info0=target_val,
          aff0=q.qualname_of(cl.addr)
        ))

    for cl in q.invocations(InvocationPattern('invoke-direct', r'java/io/File(Writer|OutputStream)?;-><init>\(Ljava/lang/String;\)')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        target_val = DataFlow(q).solved_constant_data_in_invocation(cl, 0)

        if re.search(r'debug|log|info|report|screen|err|tomb|drop', target_val):
          if not re.search(r'^/proc/|^/sys/', target_val):
            self._helper.raise_issue(self._helper.build_issue(
              sigid=self._id,
              cfd='tentative',
              cvss=self._cvss1,
              title=self._summary1,
              summary=self._synopsis1,
              info0=target_val,
              aff0=q.qualname_of(cl.addr)
            ))
      except DataFlow.NoSuchValueError:
        target_val = '(unknown name)'

class SecurityInsecureRootedDetector(SignatureMixin):
  _id = 'security-insecure-rooted'
  _cvss = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'

  _pat_path = r'^/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+'
  _pat_detect = r'Sup(?!p)|/su(?!pp)|xbin|sbin|root'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecurityInsecureRootedDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects insecure rooted device probes')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()

    attestations: Dict[str, Any] = dict()
    path_based_detection_attempt: Dict[str, Any] = dict()
    confidence: Dict[str, IssueConfidence] = dict()

    for cl in q.invocations(InvocationPattern('invoke-', r'Lcom/google/android/gms/safetynet/SafetyNetClient;->attest\(\[BLjava/lang/String;\)')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      # XXX: crude detection
      verdict_accesses = list(q.consts_in_class(cl.addr, InvocationPattern('const-string', r'ctsProfileMatch|basicIntegrity')))
      if verdict_accesses and qn is not None:
        attestations[qn] = verdict_accesses

    for cl in q.consts(InvocationPattern('const-string', self._pat_path)):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      if qn is not None:
        attempts = self._get_attempts(self._an.get_param(cl, 1).v)
        if attempts:
          confidence[qn] = 'firm'
          if qn not in path_based_detection_attempt:
            path_based_detection_attempt[qn] = attempts
          else:
            path_based_detection_attempt[qn].update(attempts)
    for name, val in context.string_resources():
      qn = f'R.string.{name}'
      attempts = self._get_attempts(val)
      if attempts:
        confidence[qn] = 'tentative'
        if qn not in path_based_detection_attempt:
          path_based_detection_attempt[qn] = attempts
        else:
          path_based_detection_attempt[qn].update(attempts)

    # TBD: fussy match
    for qn,vals in path_based_detection_attempt.items():
      if qn not in attestations:
        self._helper.raise_issue(self._helper.build_issue(sigid=self._id, cvss=self._cvss, title='manual root detections without remote attestations', info0=','.join(vals), aff0=qn))
    for qn,verdicts in attestations.items():
      if qn not in path_based_detection_attempt:
        self._helper.raise_issue(self._helper.build_issue(sigid=self._id, cvss=self._cvss, title='remote attestations without manual root detections', info0=','.join(verdicts), aff0=qn))

  def _get_attempts(self, x: str) -> Set[str]:
    o: Set[str] = set()
    for m in re.finditer(self._pat_path, x):
      v = m.group(0)
      if re.search(self._pat_detect, v):
        o.add(v)
    return o

class SecuritySharedPreferencesDetector(SignatureMixin):
  _id = 'security-sharedpref'
  _cvss = 'CVSS:3.0/AV:L/AC:H/PR:N/UI:N/S:U/C:N/I:N/A:N/'
  _summary = 'detected SharedPreference access'
  _synopsis = 'The application is using SharedPreferences. This is purely informational; Using the subsystem alone does not constitute a security issue.'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return SecuritySharedPreferencesDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects SharedPreferences access')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for cl in q.invocations(InvocationPattern('invoke-interface', r'Landroid/content/SharedPreferences;->get(Boolean|Float|Int|String|StringSet)\(Ljava/lang/String;')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        target_val = DataFlow(q).solved_constant_data_in_invocation(cl, 0)
      except DataFlow.NoSuchValueError:
        target_val = '(unknown name)'

      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cfd='certain',
        cvss=self._cvss,
        title=self._summary,
        summary=self._synopsis,
        info0=target_val,
        info1='read',
        aff0=q.qualname_of(cl.addr)
      ))

    for cl in q.invocations(InvocationPattern('invoke-interface', r'Landroid/content/SharedPreferences\$Editor;->put(Boolean|Float|Int|String|StringSet)\(Ljava/lang/String;')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        target_val = DataFlow(q).solved_constant_data_in_invocation(cl, 0)
      except DataFlow.NoSuchValueError:
        target_val = '(unknown name)'

      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cfd='certain',
        cvss=self._cvss,
        title=self._summary,
        summary=self._synopsis,
        info0=target_val,
        info1='write',
        aff0=q.qualname_of(cl.addr)
      ))

    for cl in q.invocations(InvocationPattern('invoke-interface', r'Landroid/content/SharedPreferences/Editor;->remove\(Ljava/lang/String;')):
      qn = q.qualname_of(cl.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        target_val = DataFlow(q).solved_constant_data_in_invocation(cl, 0)
      except DataFlow.NoSuchValueError:
        target_val = '(unknown name)'

      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cfd='certain',
        cvss=self._cvss,
        title=self._summary,
        summary=self._synopsis,
        info0=target_val,
        info1='delete',
        aff0=q.qualname_of(cl.addr)
      ))



from __future__ import annotations
from typing import TYPE_CHECKING

import re
from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from trueseeing.api import CommandHelper, Command, CommandMap

class SearchCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return SearchCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      '/f':dict(e=self._search_file, n='/f [pat]', d='search files those names matching pattern'),
      '/s':dict(e=self._search_string, n='/s pat [filepat]', d='search files for string'),
    }

  async def _search_file(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if args:
      pat = args.popleft()
    else:
      pat = '.'

    context = await self._helper.get_context().analyze(level=1)
    level = context.get_analysis_level()
    if level < 3:
      ui.warn('detected analysis level: {} ({}) -- try analyzing fully (\'aa\') to maximize coverage'.format(level, self._helper.decode_analysis_level(level)))
    for path in context.store().query().file_find(pat=pat, regex=True):
      ui.info(f'{path}')

  async def _search_string(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need pattern')

    pat = args.popleft()

    if args:
      mask = args.popleft()
    else:
      mask = None

    ui.info('searching: {pat}{mask}'.format(pat=pat, mask=' [file mask: {}]'.format(mask) if mask else ''))

    context = await self._helper.get_context().analyze(level=1)
    level = context.get_analysis_level()
    if level < 3:
      ui.warn('detected analysis level: {} ({}) -- try analyzing fully (\'aa\') to maximize coverage'.format(level, self._helper.decode_analysis_level(level)))
    for path, blob in context.store().query().file_enum(mask):
      if self._looks_binary(blob):
        for m0 in re.finditer(pat.encode('latin1'), blob):
          ui.info('{}:+{:08x}: {!r}'.format(path, m0.start(), m0.group(0)))
      else:
        from io import BytesIO
        for nr, t in enumerate(l.rstrip() for l in BytesIO(blob)):
          m = re.search(pat.encode('utf-8'), t)
          if m:
            ui.info('{}:{:d}:{:d}: {}'.format(path, nr+1, m.start(), t.decode('utf-8', 'replace')))

  def _looks_binary(self, b: bytes) -> bool:
    return b'\x00' in b



from __future__ import annotations
from typing import TYPE_CHECKING

import re
from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui
from trueseeing.core.android.device import AndroidDevice

if TYPE_CHECKING:
  from typing import Optional, Tuple, Literal, AsyncIterator, Dict
  from trueseeing.api import CommandHelper, Command, CommandMap
  from trueseeing.core.android.context import APKContext

  UIPatternType = Literal['re', 'xpath']

class ReconCommand(CommandMixin):
  _target_only: bool
  _watch_logcat: Optional[bytes]
  _watch_intent: Optional[bytes]
  _watch_fs: Optional[bytes]
  _watch_ui: Optional[Tuple[str, UIPatternType]]
  _watch_ui_outfn: Optional[str]

  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper
    self._target_pid = None
    self._target_only = False
    self._watch_logcat = None
    self._watch_intent = None
    self._watch_ui = None
    self._watch_ui_outfn = None
    self._watch_fs = None

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return ReconCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      '!!':dict(e=self._recon_shell, n='!!', d='run shell on device'),
      'rwl':dict(e=self._recon_watch_logcat, n='rwl[!] [pat]', d='recon: watch logcat (!: system-wide)'),
      'rwl!':dict(e=self._recon_watch_logcat),
      'rwf':dict(e=self._recon_watch_fs, n='rwf', d='recon: watch filesystem'),
      'rwt':dict(e=self._recon_watch_intent, n='rwt[!] [pat]', d='recon: watch intent'),
      'rwt!':dict(e=self._recon_watch_intent),
      'rwu':dict(e=self._recon_watch_ui, n='rwu[!] [pat|xp:xpath] [output.xml]', d='recon: watch device UI'),
      'rwu!':dict(e=self._recon_watch_ui),
      'rwx':dict(e=self._recon_watch_start, n='rwx', d='recon: start watching'),
      'rp':dict(e=self._recon_list_packages, n='rp', d='recon: list installed packages'),
      'ru':dict(e=self._recon_dump_ui, n='ru [output.xml]', d='recon: dump device UI'),
    }

  def _get_apk_context(self) -> APKContext:
    return self._helper.get_context().require_type('apk')

  async def _recon_shell(self, args: deque[str]) -> None:
    _ = args.popleft()

    from asyncio import create_subprocess_shell
    from trueseeing.core.android.device import AndroidDevice
    dev = AndroidDevice()
    dev.require_adb()
    cmdline = dev.get_adb_cmdline('shell')
    await (await create_subprocess_shell(cmdline)).wait()

  async def _recon_watch_logcat(self, args: deque[str]) -> None:
    cmd = args.popleft()

    if cmd.endswith('!'):
      self._target_only = False
    else:
      self._target_only = True
      _ = self._helper.require_target()

    if not args:
      self._watch_logcat = None
    else:
      pat = args.popleft()
      self._watch_logcat = pat.encode('latin1')

    if self._watch_logcat:
      ui.success('logcat watch enabled: {}'.format(self._watch_logcat.decode('latin1')))
    else:
      ui.success('logcat watch disabled')

  async def _recon_watch_intent(self, args: deque[str]) -> None:
    _ = args.popleft()

    if not args:
      self._watch_intent = None
    else:
      pat = args.popleft()
      self._watch_intent = pat.encode('latin1')

    if self._watch_intent:
      ui.success('intent watch enabled: {}'.format(self._watch_intent.decode('latin1')))
    else:
      ui.success('intent watch disabled')

  async def _recon_watch_fs(self, args: deque[str]) -> None:
    _ = args.popleft()

    if not args:
      self._watch_fs = None
    else:
      pat = args.popleft()
      self._watch_fs = pat.encode('latin1')

    if self._watch_fs:
      ui.success('filesystem watch enabled: {}'.format(self._watch_fs.decode('latin1')))
    else:
      ui.success('filesystem watch disabled')

  async def _recon_watch_ui(self, args: deque[str]) -> None:
    cmd = args.popleft()

    if not args:
      self._watch_ui = None
    else:
      pat = args.popleft()
      if pat.startswith('xp:'):
        self._watch_ui = pat[3:], 'xpath'
      else:
        self._watch_ui = pat, 're'

      if args:
        import os
        outfn = args.popleft()
        if os.path.exists(outfn) and not cmd.endswith('!'):
          ui.fatal('outfile exists; force (!) to overwrite')
        self._outfn_ui = outfn

    if self._watch_ui:
      ui.success('ui watch enabled: {} [{}]'.format(self._watch_ui[0], self._watch_ui[1]))
    else:
      ui.success('ui watch disabled')

  async def _recon_watch_start(self, args: deque[str]) -> None:
    if not (self._watch_logcat or self._watch_intent or self._watch_ui or self._watch_fs):
      ui.fatal('nothing to watch (try d* beforehand)')

    dev = AndroidDevice()
    pkg: Optional[str] = None
    if self._target_only:
      ctx = self._get_apk_context()
      pkg = ctx.get_package_name()

    async def _log() -> None:
      pid: Optional[int] = None
      if self._watch_logcat or self._watch_intent:
        if pkg and not pid:
          d = await dev.invoke_adb('shell ps')
          m = re.search(r'^[0-9a-zA-Z_]+ +([0-9]+) .*{}$'.format(pkg).encode(), d.encode(), re.MULTILINE)
          if m:
            pid = int(m.group(1))
            ui.info(f'detected target at pid: {pid}')

        async for l in dev.invoke_adb_streaming('logcat -T1'):
          l = l.rstrip()
          if pkg and not pid:
            m = re.search(r' ([0-9]+):{}'.format(pkg).encode(), l)
            if m:
              pid = int(m.group(1))
              ui.info(f'detected target at pid: {pid}')
          if self._watch_logcat:
            if self._target_only:
              if not (pid and re.search(r'\.[0-9]+? +{} +'.format(pid).encode(), l)):
                continue
            if re.search(self._watch_logcat, l):
              ui.info('log: {}'.format(l.decode('latin1')))
          if self._watch_intent and b'intent' in l:
            if re.search(self._watch_intent, l):
              ui.info('intent: {}'.format(l.decode('latin1')))
          m = re.search('([0-9]+) +?[0-9]+? I Frida +?: Listening on (.*)'.encode(), l)
          if m:
            gad_pid, gad_port = int(m.group(1)), m.group(2).decode()
            ui.success('detected frida-gadget listening on {} for pid {}{}'.format(gad_port, gad_pid, ' [target]' if pid == gad_pid else ''))

    async def _ui() -> None:
      import lxml.etree as ET
      import re

      if self._watch_ui:
        pat, typ = self._watch_ui

        async for dom in self._dump_ui_cont():
          matched = False
          if typ == 'xpath':
            r = ET.fromstring(dom)
            e = r.xpath(pat)
            if e:
              matched = True
              ui.info('ui: {} [{}]: found {}{}'.format(pat, typ, len(e), ' (dumped)' if self._watch_ui_outfn else ''))
          else:
            m = re.search(pat, dom)
            if m:
              matched = True
              ui.info('ui: {} [{}]: found {}'.format(pat, typ, ' (dumped)' if self._watch_ui_outfn else ''))

          if matched:
            if self._watch_ui_outfn:
              with open(self._watch_ui_outfn, 'w') as f:
                f.write(dom)

    async def _fs() -> None:
      if self._watch_fs:
        async for mode, path, md in self._watch_fs_mod_cont(delay=2.0):
          if re.search(self._watch_fs, path):
            ui.info('fs: {}: {} {}'.format(mode.decode(), md.decode(), path.decode()))

    try:
      from asyncio import gather
      ui.info('watching device (C-c to stop)')
      await gather(_log(), _ui(), _fs(), return_exceptions=True)
    except KeyboardInterrupt:
      pass

  async def _recon_list_packages(self, args: deque[str]) -> None:
    _ = args.popleft()

    ui.info('listing packages')

    import time
    import re
    from trueseeing.core.android.device import AndroidDevice

    at = time.time()
    nr = 0
    for m in re.finditer(r'^package:(.*)', await AndroidDevice().invoke_adb('shell pm list package'), re.MULTILINE):
      p = m.group(1)
      ui.info(p)
      nr += 1
    ui.success('done, {nr} packages found ({t:.02f} sec.)'.format(nr=nr, t=(time.time() - at)))

  async def _recon_dump_ui(self, args: deque[str]) -> None:
    outfn: Optional[str] = None

    cmd = args.popleft()

    if args:
      import os
      outfn = args.popleft()
      if os.path.exists(outfn) and not cmd.endswith('!'):
        ui.fatal('outfile exists; force (!) to overwrite')

    ui.info('dumping UI hierachy')

    try:
      dom = await self._dump_ui()
    except DumpFailedError as e:
      ui.fatal(f'dump failed: {e}')

    if outfn is None:
      ui.stdout(dom)
    else:
      with open(outfn, 'w') as f:
        f.write(dom)
    ui.success('done')

  async def _dump_ui(self) -> str:
    from subprocess import CalledProcessError
    dev = AndroidDevice()
    try:
      msg = await dev.invoke_adb('shell uiautomator dump')
      m = re.search(r'dumped to: (/.*)', msg)
      if not m:
        raise DumpFailedError(msg)
      tmpfn = m.group(1)
      return await dev.invoke_adb(f"shell 'cat {tmpfn}; rm {tmpfn}'")
    except CalledProcessError as e:
      raise DumpFailedError(e)

  async def _dump_ui_cont(self, /, delay: float = 1.0) -> AsyncIterator[str]:
    from subprocess import CalledProcessError
    dev = AndroidDevice()
    try:
      async for msg in dev.invoke_adb_streaming(r"shell 'while (true) do uiautomator dump /sdcard/dump-$(date +%s).xml; sleep {delay}; done'".format(delay=delay)):
        m = re.search(r'dumped to: (/.*)', msg.decode())
        if m:
          tmpfn = m.group(1)
          yield await dev.invoke_adb(f"shell 'cat {tmpfn}; rm {tmpfn}'")
    except CalledProcessError as e:
      raise DumpFailedError(e)

  async def _watch_fs_mod_cont(self, /, delay: float = 1.0) -> AsyncIterator[Tuple[bytes, bytes, bytes]]:
    from subprocess import CalledProcessError

    seen0: Dict[bytes, bytes] = dict()
    seen1: Dict[bytes, bytes] = dict()

    dev = AndroidDevice()
    try:
      async for msg in dev.invoke_adb_streaming(r"shell 'while (true) do (find -H /storage -mtime -{thres}s -print0 | xargs -0 ls -nlld); echo \"*\"; sleep {delay}; done'".format(delay=delay, thres=int(delay*3.))):
        msg = msg.rstrip()
        if not msg.startswith(b'*'):
          m = re.fullmatch(r'(.*00) (.+?)'.encode('latin1'), msg)
          if m:
            fn = m.group(2)
            metadata = m.group(1)
            seen0[fn] = metadata
        else:
          seen0set = frozenset(seen0.keys())
          seen1set = frozenset(seen1.keys())

          for k in seen0set - seen1set:
            yield b'add', k, seen0[k]

          for k in seen0set & seen1set:
            if seen0[k] != seen1[k]:
              yield b'mod', k, seen0[k]

          seen1.update(seen0)
          seen0.clear()
    except CalledProcessError as e:
      raise DumpFailedError(e)

class DumpFailedError(Exception):
  pass



from __future__ import annotations
from typing import TYPE_CHECKING

from trueseeing.api import Command

if TYPE_CHECKING:
  from trueseeing.api import CommandMap, CommandPatternMap, ModifierMap, OptionMap, ConfigMap

class CommandMixin(Command):
  def get_commands(self) -> CommandMap:
    return dict()
  def get_command_patterns(self) -> CommandPatternMap:
    return dict()
  def get_modifiers(self) -> ModifierMap:
    return dict()
  def get_configs(self) -> ConfigMap:
    return dict()
  def get_options(self) -> OptionMap:
    return dict()
